"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const constant_1 = require("./constant");
const validate_1 = require("./validate");
// import { Util } from './util'
const query_1 = require("./serializer/query");
const update_1 = require("./serializer/update");
const constant_2 = require("./constant");
const util_1 = require("./util");
const datatype_1 = require("./serializer/datatype");
/**
 * Db query
 */
class Query {
    /**
     * @param db            - db reference
     * @param coll          - collection name
     * @param fieldFilters  - query condition
     * @param fieldOrders   - order by condition
     * @param queryOptions  - query options
     */
    constructor(db, coll, fieldFilters, fieldOrders, queryOptions, withs) {
        this._db = db;
        this._coll = coll;
        this._fieldFilters = fieldFilters;
        this._fieldOrders = fieldOrders || [];
        this._queryOptions = queryOptions || {};
        this._withs = withs || [];
        this._request = this._db.request;
    }
    /**
     * 查询条件
     *
     * @param query
     */
    where(query) {
        // query校验 1. 必填对象类型  2. value 不可均为 undefiend
        if (Object.prototype.toString.call(query).slice(8, -1) !== 'Object') {
            throw Error(constant_2.ErrorCode.QueryParamTypeError);
        }
        const keys = Object.keys(query);
        const checkFlag = keys.some((item) => {
            return query[item] !== undefined;
        });
        if (keys.length && !checkFlag) {
            throw Error(constant_2.ErrorCode.QueryParamValueError);
        }
        const _query = query_1.QuerySerializer.encode(query);
        return new Query(this._db, this._coll, _query, this._fieldOrders, this._queryOptions, this._withs);
    }
    /**
     * 设置排序方式
     *
     * @param fieldPath     - 字段路径
     * @param directionStr  - 排序方式
     */
    orderBy(fieldPath, directionStr) {
        validate_1.Validate.isFieldPath(fieldPath);
        validate_1.Validate.isFieldOrder(directionStr);
        const newOrder = {
            field: fieldPath,
            direction: directionStr,
        };
        const combinedOrders = this._fieldOrders.concat(newOrder);
        return new Query(this._db, this._coll, this._fieldFilters, combinedOrders, this._queryOptions, this._withs);
    }
    /**
     * 添加 一对多 子查询条件
     * @param param {WithParam}
     * @returns Query
     */
    with(param) {
        var _a, _b;
        const newWith = {
            query: param.query,
            foreignField: param.foreignField,
            localField: param.localField,
            as: (_a = param.as) !== null && _a !== void 0 ? _a : param.query._coll,
            one: (_b = param.one) !== null && _b !== void 0 ? _b : false,
        };
        const combinedWiths = this._withs.concat(newWith);
        return new Query(this._db, this._coll, this._fieldFilters, this._fieldOrders, this._queryOptions, combinedWiths);
    }
    /**
     * 添加 一对一 子查询条件
     * @param param {WithParam}
     * @returns Query
     */
    withOne(param) {
        var _a;
        const newWith = {
            query: param.query,
            foreignField: param.foreignField,
            localField: param.localField,
            as: (_a = param.as) !== null && _a !== void 0 ? _a : param.query._coll,
            one: true,
        };
        const combinedWiths = this._withs.concat(newWith);
        return new Query(this._db, this._coll, this._fieldFilters, this._fieldOrders, this._queryOptions, combinedWiths);
    }
    /**
     * 指定要返回的字段
     *
     * @param projection
     */
    field(projection) {
        let formatted = {};
        if (projection instanceof Array) {
            const result = {};
            for (const k of projection) {
                result[k] = 1;
            }
            formatted = result;
        }
        else {
            for (const k in projection) {
                if (projection[k]) {
                    if (typeof projection[k] !== 'object') {
                        formatted[k] = 1;
                    }
                }
                else {
                    formatted[k] = 0;
                }
            }
        }
        const option = Object.assign({}, this._queryOptions);
        option.projection = formatted;
        return new Query(this._db, this._coll, this._fieldFilters, this._fieldOrders, option, this._withs);
    }
    /**
     * 设置查询条数
     *
     * @param limit - 限制条数，当前限制一次请求获取数据条数不得大于 1000
     */
    limit(limit) {
        validate_1.Validate.isInteger('limit', limit);
        const option = Object.assign({}, this._queryOptions);
        option.limit = limit;
        return new Query(this._db, this._coll, this._fieldFilters, this._fieldOrders, option, this._withs);
    }
    /**
     * 设置偏移量
     *
     * @param offset - 偏移量
     */
    skip(offset) {
        validate_1.Validate.isInteger('offset', offset);
        const option = Object.assign({}, this._queryOptions);
        option.offset = offset;
        return new Query(this._db, this._coll, this._fieldFilters, this._fieldOrders, option, this._withs);
    }
    /**
     * 设置分页查询
     * @param options { current: number, size: number} `current` 是页码，默认为 `1`, `size` 是每页大小, 默认为 10
     */
    page(options) {
        const current = (options === null || options === void 0 ? void 0 : options.current) || 1;
        const size = (options === null || options === void 0 ? void 0 : options.size) || 10;
        const query = this.skip((current - 1) * size).limit(size);
        query._queryOptions.count = true;
        return query;
    }
    /**
     * 克隆
     * @returns Query
     */
    clone() {
        return new Query(this._db, this._coll, this._fieldFilters, this._fieldOrders, this._queryOptions, this._withs);
    }
    /**
     * 发起请求获取文档列表
     *
     * - 默认 `limit` 为 100
     * - 可以把通过 `orderBy()`、`where()`、`skip()`、`limit()`设置的数据添加请求参数上
     */
    async get() {
        var _a;
        if ((_a = this._withs) === null || _a === void 0 ? void 0 : _a.length) {
            return await this.internalMerge();
        }
        else {
            return await this.internalGet();
        }
    }
    /**
     * 发起请求获取一个文档
     * @param options
     * @returns
     */
    async getOne() {
        const res = await this.limit(1).get();
        if (res.error) {
            return res;
        }
        if (!res.data.length) {
            return {
                ok: true,
                data: null,
                requestId: res.requestId,
            };
        }
        return {
            ok: true,
            data: res.data[0],
            requestId: res.requestId,
        };
    }
    /**
     * [该接口已废弃，直接使用 `get()` 代替]
     * 发起请求获取文档列表，当使用 with 条件时使用
     *
     * @deprecated
     *
     * 1. 调用 get() 执行主查询
     * 2. 结合主查询的结果，使用 in 执行子表查询
     * 3. 合并主表 & 子表的结果，即聚合
     * 4. intersection 可指定是否取两个结果集的交集，缺省则以主表结果为主
     */
    async merge(options) {
        const res = await this.internalMerge(options);
        return res;
    }
    /**
     * 获取总数
     */
    async count() {
        const param = this.buildQueryParam();
        const res = await this.send(constant_1.ActionType.count, param);
        if (res.error) {
            return {
                requestId: res.requestId,
                ok: false,
                error: res.error,
                total: undefined,
                code: res.code,
            };
        }
        return {
            requestId: res.requestId,
            total: res.data.total,
            ok: true,
        };
    }
    /**
     * 发起请求批量更新文档
     *
     * @param data 数据
     */
    async update(data, options) {
        var _a, _b, _c, _d;
        if (!data || typeof data !== 'object' || 0 === ((_a = Object.keys(data)) === null || _a === void 0 ? void 0 : _a.length)) {
            throw new Error('data cannot be empty object');
        }
        if (data.hasOwnProperty('_id')) {
            throw new Error('can not update the `_id` field');
        }
        const param = this.buildQueryParam();
        param.multi = (_b = options === null || options === void 0 ? void 0 : options.multi) !== null && _b !== void 0 ? _b : false;
        param.merge = (_c = options === null || options === void 0 ? void 0 : options.merge) !== null && _c !== void 0 ? _c : true;
        param.upsert = (_d = options === null || options === void 0 ? void 0 : options.upsert) !== null && _d !== void 0 ? _d : false;
        if (param.merge) {
            param.data = update_1.UpdateSerializer.encode(data);
        }
        else {
            param.data = datatype_1.serialize(data);
        }
        const res = await this.send(constant_1.ActionType.update, param);
        if (res.error) {
            return {
                requestId: res.requestId,
                error: res.error,
                ok: false,
                code: res.code,
                updated: undefined,
                matched: undefined,
                upsertId: undefined,
            };
        }
        return {
            requestId: res.requestId,
            updated: res.data.updated,
            matched: res.data.matched,
            upsertId: res.data.upsert_id,
            ok: true,
        };
    }
    /**
     * 条件删除文档
     */
    async remove(options) {
        var _a, _b;
        if (Object.keys(this._queryOptions).length > 0) {
            console.warn('`offset`, `limit` and `projection` are not supported in remove() operation');
        }
        if (((_a = this._fieldOrders) === null || _a === void 0 ? void 0 : _a.length) > 0) {
            console.warn('`orderBy` is not supported in remove() operation');
        }
        const param = this.buildQueryParam();
        param.multi = (_b = options === null || options === void 0 ? void 0 : options.multi) !== null && _b !== void 0 ? _b : false;
        const res = await this.send(constant_1.ActionType.remove, param);
        if (res.error) {
            return {
                requestId: res.requestId,
                error: res.error,
                ok: false,
                deleted: undefined,
                code: res.code,
            };
        }
        return {
            requestId: res.requestId,
            deleted: res.data.deleted,
            ok: true,
        };
    }
    /**
     * Build query param
     * @returns
     */
    buildQueryParam() {
        var _a;
        const param = {
            collectionName: this._coll,
        };
        if (this._fieldFilters) {
            param.query = this._fieldFilters;
        }
        if ((_a = this._fieldOrders) === null || _a === void 0 ? void 0 : _a.length) {
            param.order = [...this._fieldOrders];
        }
        if (this._queryOptions.offset) {
            param.offset = this._queryOptions.offset;
        }
        if (this._queryOptions.limit) {
            param.limit =
                this._queryOptions.limit < 1000 ? this._queryOptions.limit : 1000;
        }
        else {
            param.limit = 100;
        }
        if (this._queryOptions.projection) {
            param.projection = this._queryOptions.projection;
        }
        if (this._queryOptions.count) {
            param.count = this._queryOptions.count;
        }
        return param;
    }
    /**
     * 发起请求获取文档列表
     */
    async internalGet() {
        var _a, _b, _c;
        const param = this.buildQueryParam();
        const res = await this.send(constant_1.ActionType.query, param);
        if (res.error) {
            return {
                error: res.error,
                data: res.data,
                requestId: res.requestId,
                ok: false,
                code: res.code,
            };
        }
        const documents = util_1.Util.formatResDocumentData(res.data.list);
        const result = {
            data: documents,
            requestId: res.requestId,
            ok: true,
        };
        if (res.total)
            result.total = (_a = res.data) === null || _a === void 0 ? void 0 : _a.total;
        if (res.limit)
            result.limit = (_b = res.data) === null || _b === void 0 ? void 0 : _b.limit;
        if (res.offset)
            result.offset = (_c = res.data) === null || _c === void 0 ? void 0 : _c.offset;
        return result;
    }
    /**
     * 发起请求获取文档列表，当使用 with 条件时使用
     *
     * 1. 调用 internalGet() 执行主查询
     * 2. 结合主查询的结果，使用 in 执行子表查询
     * 3. 合并主表 & 子表的结果，即聚合
     * 4. intersection 可指定是否取两个结果集的交集，缺省则以主表结果为主
     */
    async internalMerge(options) {
        var _a;
        options = options !== null && options !== void 0 ? options : {};
        const intersection = (_a = options === null || options === void 0 ? void 0 : options.intersection) !== null && _a !== void 0 ? _a : false;
        // 调用 get() 执行主查询
        const res = await this.internalGet();
        if (!res.ok) {
            return res;
        }
        // 针对每一个 WithParam 做合并处理
        for (const _with of this._withs) {
            const { query, localField, foreignField, as, one } = _with;
            const localValues = res.data.map((localData) => localData[localField]);
            // 处理子查询
            const q = query.clone();
            if (!q._fieldFilters) {
                q._fieldFilters = {};
            }
            q._fieldFilters[foreignField] = { $in: localValues };
            // 执行子查询
            let r_sub;
            if (q._withs.length) {
                r_sub = await q.merge(); // 如果子查询也使用了 with/withOne，则使用 merge() 查询
            }
            else {
                r_sub = await q.get();
            }
            if (!r_sub.ok) {
                return r_sub;
            }
            // 按照 localField -> foreignField 的连接关系将子查询结果聚合：
            // 1. 构建 { [value of `foreignField`]: [subQueryData] } 映射表
            const _map = {};
            for (const sub of r_sub.data) {
                const key = sub[foreignField]; // 将子表结果的连接键的值做为映射表的 key
                if (one) {
                    _map[key] = sub;
                }
                else {
                    _map[key] = _map[key] || [];
                    _map[key].push(sub); // 将子表结果放入映射表
                }
            }
            // 2. 将聚合结果合并入主表结果集中
            const results = [];
            for (const m of res.data) {
                // 此处主表结果中的 [value of `localField`] 与 上面子表结果中的 [value of `foreignField`] 应该是一致的
                const key = m[localField];
                m[as] = _map[key];
                // 如果取交集且子表结果无对应数据，则丢弃此条数据
                if (intersection && !_map[key]) {
                    continue;
                }
                results.push(m);
            }
            res.data = results;
        }
        return res;
    }
    /**
     * Send query request
     * @param action
     * @param param
     * @returns
     */
    async send(action, param) {
        return await this._request.send(action, param);
    }
}
exports.Query = Query;
